"""Authentication Blueprint"""
import functools

from flask import (
    Blueprint,
    Response,
    flash,
    g,
    redirect,
    render_template,
    request,
    session,
    url_for,
)
from sqlalchemy.exc import IntegrityError
from werkzeug.security import check_password_hash, generate_password_hash

from bcitflex.db import DBSession
from bcitflex.model import User

bp = Blueprint("auth", __name__, url_prefix="/auth")


def redirect_dest(fallback: str) -> Response:
    """Redirect to the next page after login."""
    dest = request.args.get("next")
    if dest is None:
        return redirect(fallback)
    return redirect(url_for(dest))


@bp.route("/register", methods=("GET", "POST"))
def register():
    if request.method == "POST":
        username = request.form["username"]
        password = request.form["password"]
        db = DBSession()
        error = None

        if not username:
            error = "Username is required."
        elif not password:
            error = "Password is required."

        if error is None:
            try:
                new_user = User(
                    username=username, password=generate_password_hash(password)
                )
                with db.begin():
                    db.add(new_user)
            except IntegrityError:
                error = f"User {username} is already registered."
            else:
                return redirect(url_for("auth.login"))

        flash(error)

    return render_template("auth/register.html")


@bp.route("/login", methods=("GET", "POST"))
def login():
    if request.method == "POST":
        username = request.form["username"]
        password = request.form["password"]
        db = DBSession()
        error = None
        user = User.get_by_unique(db, username)

        if user is None:
            error = "Incorrect username."
        elif not check_password_hash(user["password"], password):
            error = "Incorrect password."

        if error is None:
            session.clear()
            session["user_id"] = user["id"]
            session["programs"] = user.preference.programs
            return redirect_dest(url_for("index"))

        flash(error)

    return render_template("auth/login.html")


@bp.before_app_request
def load_logged_in_user():
    user_id = session.get("user_id")

    if user_id is None:
        g.user = None
    else:
        g.user = DBSession().get(User, user_id)


@bp.route("/logout")
def logout():
    session.clear()
    return redirect(url_for("index"))


def login_required(view):
    @functools.wraps(view)
    def wrapped_view(**kwargs):
        if g.user is None:
            return redirect(url_for("auth.login", next=request.endpoint))

        return view(**kwargs)

    return wrapped_view
